import urllib.request
from bs4 import BeautifulSoup
import spacy
import neuralcoref
from nltk.tokenize import sent_tokenize


class ExtractedSentences:
    """
    ExtractedSentences

    Class to preprocess the sentences. Preprocessing happens in two ways
    1. The corefrence is resolved using scapy and neuralcoref
    2. The text is tokenized into separate sentences.

    Arguments:
        text: String; The transcipt generally.
    """

    def __init__(self, text: str):
        """
        __init__

        Arguments:
            text: str
        """

        self.text = text
        self.nlp = spacy.load('en_core_web_lg')
        neuralcoref.add_to_pipe(self.nlp)

    def __resolve_coreference(self):
        _doc = self.nlp(self.text)
        self.resolved_text = _doc._.coref_resolved

    def __tokenize_sentences(self):
        self.tokenized_sentences = sent_tokenize(self.resolved_text)

    def get_processed_sentences(self):
        """
        Function returns the processed sentences list

        Returns:
            List of strings
        """

        self.__resolve_coreference()
        self.__tokenize_sentences()

        return self.tokenized_sentences



from rake_nltk import Rake

class Tags:
    """
    Tags: Class encapsulates the rake nltk library

    Methods:
        get_ranked_phrases( String )
    """

    def __init__(self, text: str):
        """
        __init__

        Arguments:
            text: String; The text to be summarized.
        """

        self.rake = Rake()
        self.rake.extract_keywords_from_text(text)

    def get_ranked_phrases(self):
        """
        get_ranked_phrases()

        Arguments:
            --NULL--

        Returns:
            List: sorted list of keywords in the text.
        """

        return self.rake.get_ranked_phrases()



from .Preprocessing import Preprocess
from .QuestionGeneration import TrueFlase, FillInTheBlanks, MultipleCorrect, IntergerType, SequenceRearrange
from . import Selection
import json
import random
import os

class Brain:

    def __init__(self, text: str, token_url = None, token_id = None, fill_in_the_blanks = True, true_or_false = True, multiple_corrent = True, interger_type = True):
        self.text = text
        self.pr = Preprocess.Preprocess(self.text, token_url, token_id).get_processed_sentences()
        self.li = {}

        if true_or_false:
            self.tf = TrueFlase.TrueFalse(self.pr)
            self.li[1] = self.tf.questions()


        if fill_in_the_blanks:
            self.fb = FillInTheBlanks.FillInTheBlanks(self.pr)
            self.li[2] = self.fb.questions()

        if multiple_corrent:
            self.mc = MultipleCorrect.MultipleCorrect(self.pr)
            self.li[3] = self.mc.questions()

        if interger_type:
            self.ind = IntergerType.IntergerType(self.pr)
            self.li[4] = self.ind.questions()

        self.fg = Selection.Selection(self.pr, self.li, True).get_final_question()

    def generate_question(self):
        return self.fg



from word2number import w2n
import copy


class IntergerType:
    """
    IntergerType

    Class to generate IntergerType questions, by replacing numeric values in
    the text, using NLTK and word2number python package

    Arguments:
        processed_sentences: dict
    """

    def __init__(self, processed_sentences: dict):
        """
        __init__

        Arguments:
            processed_sentences
        """
        self.processed_sentences = processed_sentences
        self.question = []


    def __generate_question(self):
        """
        generate question

        Traverse through the list of numeric words and adjectives
        """
        for i in self.processed_sentences['processed-sentences']:
            if 'CD' in i:
                for _jk in i['CD']:
                    _ques = {}
                    _ques['question'] = i['sentence'].replace(str(_jk), "____")
                    try:
                        _ques['answer'] = w2n.word_to_num(str(_jk))

                        if not isinstance(_ques['answer'], int) or _ques['answer'] == 1:
                            raise Exception

                        _ques['score'] = 1
                        _ques['type'] = 4
                        self.question.append(_ques)
                    except:
                        continue
            elif 'JJ' in i:
                for _jk in i['JJ']:
                    _ques = {}
                    _ques['question'] = i['sentence'].replace(str(_jk), "____")
                    try:
                        _ques['answer'] = w2n.word_to_num(str(_jk))

                        if not isinstance(_ques['answer'], int) or _ques['answer'] == 1:
                            raise Exception

                        _ques['score'] = 1
                        _ques['type'] = 4
                        self.question.append(_ques)
                    except:
                        continue



    def questions(self):
        """
        Returns question in the following format
        {
        question: <>,
        answer: <int>,
        score: 1,
        type: 4
        }

        """
        self.__generate_question()
        return self.question



import language_tool_python

class Grammar:
    """
    Grammar

    Class rectifies grammatical mistakes and spelling errors using
    language_tool_python.

    Arguments:
        list_of_sentences: List
    """

    def __init__(self, list_of_sentences: list):
        """
        __init__

        Arguments:
            list_of_sentences: List
        """

        self.list_of_sentences = list_of_sentences
        self.processed_list_of_sentences = []
        self.tool = language_tool_python.LanguageTool('en-US')

    def __rectify_grammar(self):
        for sentence in self.list_of_sentences:
            self.processed_list_of_sentences.append(self.tool.correct(sentence))

    def get_processed_sentences(self):
        """
        get_processed_sentences

        Returns:
            List of string.
        """
        self.__rectify_grammar()

        return self.processed_list_of_sentences



import random
import operator

class Selection:

    def __init__(self, preprocessed_question, list_of_questions, token=False):
        """
        list_of_questions:
        [
            questionTypeID: [<question_object>],
        ]
        """

        self.preprocessed_question = preprocessed_question
        self.list_of_questions = list_of_questions
        self.token = token
        self.final_question = []

    def __get_proper_noun(self):
        if self.token:
            return self.preprocessed_question["tag-intersection"]
        jk = set()
        for j in self.preprocessed_question["processed-sentences"]:
            if "NNP" in j:
                for k in j["NNP"]:
                    jk.add(k)
        return list(jk)

    def __select_fill_in_blanks(self):
        for i in self.list_of_questions[1]:
            if i["answer"] in self.__get_proper_noun():
                insert = True
                for k in self.final_question:
                    if i["question"] == k["question"]:
                        insert = False
                        break
                if insert:
                    self.final_question.append(i)

    def __select_true_or_false(self, type):
        if self.token:
            question_rank = {}
            for i in self.list_of_questions[type]:
                rating = 0
                for j in self.preprocessed_question["tag-intersection"]:
                    if str(j) in i["question"]:
                        rating += 1
                question_rank[i["question"]] = rating
            sorted_tuple = sorted(question_rank.items(), key=operator.itemgetter(1), reverse=True)

            for i in sorted_tuple[0:3]:
                for j in self.list_of_questions[type]:
                    if i[0] == j["question"]:
                        insert = True
                        for k in self.final_question:
                            if j["question"] == k["question"]:
                                insert = False
                                break
                        if insert:
                            j["question"] = str(j["question"])
                            j["answer"] = str(j["answer"])
                            self.final_question.append(j)

        else:
            for i in self.list_of_questions[type]:
                for j in self.preprocessed_question["tag"][0:5]:
                    if j in i["question"]:
                        j["question"] = str(j["question"])
                        j["answer"] = str(j["answer"])
                        self.final_question.append(i)


    def __select_multiple_correct(self):
        for i in self.list_of_questions[3]:
            if i["answer1"] in self.__get_proper_noun():
                if i["answer2"] in self.__get_proper_noun():
                    insert = True
                    for k in self.final_question:
                        if i["question"] == k["question"]:
                            insert = False
                            break
                    if insert:
                        self.final_question.append(i)

    def __select_relevant_question(self):

        def f(questionType):
            return {
                1: self.__select_fill_in_blanks(),
                2: self.__select_true_or_false(2),
                3: self.__select_multiple_correct(),
                4: self.__select_true_or_false(4)
            }[questionType]

        for questionType in [2, 3, 4, 5]:
            if questionType in self.list_of_questions:
                f(questionType)

        if len(self.final_question) > 2:
            random.shuffle(self.final_question)
            self.final_question = self.final_question[0:3]


    def get_final_question(self):
        self.__select_relevant_question()
        return self.final_question



from . import Tags
from . import RankRDF
from . import GetSentences
from . import Grammar
from . import POSTag


class Preprocess:
    """
    Preprocess

    Main class, acts as an entrypoint to Preprocess the text.
    Combines, integrates all the other class of the module.

    Arguments
        text: string; The text to be Preprocessed
        Tokenid: string; ID of the plugin used.
    """

    def __init__(self, text: str, toeken_url: str = None, tokenid: str = None):
        """
        __init__

        Arguments:
            text: string; The text to be Preprocessed
            Tokenid: string; ID of the plugin used.
        """

        self.text = text
        self.tokenid = tokenid
        self.token_url = toeken_url

    def __process_text(self):
        self.results = {}
        self.results["tag"] = Tags.Tags(self.text).get_ranked_phrases()

        if self.tokenid is not None:
            self.results["tag-intersection"] = RankRDF.RankRDF(
                self.results["tag"], self.token_url, self.tokenid).get_tags_intersection()

        self.results["processed-sentences"] = POSTag.GivePOSTags(
            Grammar.Grammar(
                GetSentences.ExtractedSentences(
                    self.text).get_processed_sentences()).get_processed_sentences()).get_sentences_with_tags()

    def get_processed_sentences(self):
        """
        get_processed_sentences

        Returns:
            Dictionary object of the following format
            {"tags": keywords in the text,
             "tags-intersection": Intersection of keywords from the plugin,
             "processed-sentences": List of sentences, find its schema in
              preprocess.py }
        """
        self.__process_text()

        return self.results



import en_core_web_sm
import requests
import os
import random
import nltk
from nltk.corpus import wordnet


class TrueFalse:
    """
    True and False

    Class to generate fill in the blanks type questions based on processed
    transcript. Replacing the proper nouns.

    Arguments:
        processed_transcript: dict;
    """

    def __init__(self, processed_transcript: list, tokenid: str = None):
        """
        __init__

        Arguments:
            processed_transcript: dict
        """

        self.processed_transcript = processed_transcript
        self.nlp = en_core_web_sm.load()
        self.tokenid = tokenid
        self.question = []

    @staticmethod
    def __get_replacement_from_plugin(tokenid: str, type: str):
        """
        RDF provides related replacements for a particular kind of word.
        Example: Person, organization names.
        """
        _req = requests.get(os.environ["PLUGIN_STORE_URL"] + tokenid +"/"+type)

        if _req.status_code == 200:
            return _req.get()["data"]
        return None

    @staticmethod
    def __get_antonym(word: str):
        """Provides the antonym of an adjective."""
        for syn in wordnet.synsets(word):
            for l in syn.lemmas():
                if l.antonyms():
                    return l.antonyms()[0].name()
        return None

    def __named_entity_recognition(self):
        """
        Named entity are recognized and randomly replaced
        by related words, using RDF.
        """

        for i in self.processed_transcript["processed-sentences"]:
            _doc = self.nlp(i['sentence'])
            for X in _doc.ents:
                _rep = None
                if self.tokenid:
                    _rep = self.__get_replacement_from_plugin(self.tokenid, X.label_)
                if _rep:
                    _to_flip = random.choice([0, 1])
                    _ques = {}
                    if _to_flip == 1:
                        _ques["question"] = i['sentence'].replace(X.text, _rep)
                        _ques["answer"] = False
                    else:
                        _ques["question"] = i['sentence']
                        _ques["answer"] = True
                    _ques["score"] = 1
                    _ques["type"] = 2
                    self.question.append(_ques)

    def __adjective_replacement(self):
        """Adjective are replaced with their antonyms randomly."""

        for i in self.processed_transcript["processed-sentences"]:
            if "JJ" in i:
                _ran = random.choice(i['JJ'])
                _rep = self.__get_antonym(_ran)
                if _rep:
                    _to_flip = random.choice([0, 1])
                    _ques = {}
                    if _to_flip == 1:
                        _ques["question"] = i['sentence'].replace(_ran, _rep)
                        _ques["answer"] = False
                    else:
                        _ques["question"] = i['sentence']
                        _ques["answer"] = True
                    _ques["score"] = 1
                    _ques["type"] = 2
                    self.question.append(_ques)

    def __generate_question(self):
        if self.tokenid:
            self.__named_entity_recognition()
        self.__adjective_replacement()

    def questions(self):
        """
        Returns List of question in the format

        {
            question: <>,
            answer: True/False,
            score: 1,
            type: 2
        }
        """
        self.__generate_question()
        return self.question



from bisect import bisect_right
import requests
import os

class RankRDF:
    """
    RankRDF

    Class to find the intersection of tags specified by
    the plugin and the tags extracted from the transcript

    Arguments:
        tags: List of tags extracted from the text
        tokenid: String to uniquely identify the plugin
    """

    def __init__(self, tags: list, toeken_url: str, tokenid : str):
        self.tokenid = tokenid
        self.token_url = toeken_url
        self.tags = tags
        self.final_tags_intersection = []

    def __get_tags_from_plugin(self):
        _make_request = requests.get(self.token_url + self.tokenid)

        if _make_request.status_code == 200:
            self.list_from_plugin = _make_request.json()["tags"]
        elif _make_request == 403:
            raise Exception("Unautorized request for plugin.")
        else:
            raise Exception("Invalid plugin.")

    @staticmethod
    def __binary_search(elements: list, x: str):
        i = bisect_right(elements, x)
        if i != len(elements)+1 and (elements[i-1] == x or elements[i-1] in x.split()):
            return True
        else:
            return False

    def __find_intersection(self):
        for tag in self.tags:
            if self.__binary_search(self.list_from_plugin, tag):
                self.final_tags_intersection.append(tag)

    def get_tags_intersection(self, without_intersection = True):
        """
        get_tags_intersection

        Arguments:
            --NULL--

        Return:
            List of intersection of plugin tags and extracted Tags.
        """

        self.__get_tags_from_plugin()

        if without_intersection:
            return self.list_from_plugin

        self.__find_intersection()

        return self.final_tags_intersection



import requests
import os

class SequenceRearrange:
    """
    SequenceRearrange

    Class that create SequenceRearrange question, it use the sequence defined in
    the plugin and the intersection of the tags discovered in the transcript to
    create the question

    Arguments:
        processed_transcript: dict
    """

    def __init__(self, processed_transcript: list, tokenid: str = None):
        """
        __init__

        Arguments:
            processed_transcript: dict
        """
        self.processed_transcript = processed_transcript
        self.question = []
        self.tokenid = tokenid

    def __generate_question(self):
        _req = requests.get(os.environ["PLUGIN_STORE_URL"] + self.tokenid + '/seq')

        if _req.status_code == 200:
            for i in _req.json()['seq']:
                _gh = True
                for j in i:
                    if j not in self.processed_transcript["tag-intersection"]:
                        _gh = False
                        break;
                if _gh:
                    _ques = {}
                    _inde = 1
                    for j in i:
                        _ques['option' + str(_inde)] = j
                        _inde += 1
                    _ques['score'] = len(i)
                    _ques['type'] = 5
                    self.question.append(_ques)


    def questions(self):
        """
        Returns question in the following format:
        {
        option1: <>,
        option2: <>,
        option3: <>,
        score: 3,
        type: 5
        }
        """
        if self.tokenid:
            self.__generate_question()
        return self.question



class Crossword:

    def __init__(self, processed_transcript: dict):
        self.processed_transcript = processed_transcript
        self.question = []

    def __generate_question(self):
        return 1

    def questions(self):
        self.__generate_question()
        return self.question



import random
from textblob import TextBlob
import copy

class MultipleCorrect:
    """
    MultipleCorrect

    Class to generate more than one correct question
    Arguments:
        processed_transcript: dict
    """

    def __init__(self, processed_transcript: list):
        """
        __init__

        Arguments:
            processed_transcript: dict
        """

        self.processed_transcript = processed_transcript
        self.question = []

    @staticmethod
    def __format_string(st: str):
        """Remove special characters"""
        _st = ""
        _stopwords = ",.!@#$%^&*()_+-=;<>?/"
        for i in st:
            if i not in _stopwords:
                _st += i
        return _st

    def __generate_question_noun(self):
        _test_list = []
        for i in self.processed_transcript['processed-sentences']:
            _formatted_sentence = self.__format_string(i['sentence']).split()

            n = len(_formatted_sentence)
            for _jk in range(n-2):

                if 'NN' not in i:
                    i['NN'] = []

                if 'IN' not in i:
                    i['IN'] = []

                if 'NNP' in i:
                    i['NN'] += i['NNP']
                if 'NNS' in i:
                    i['NN'] += i['NNS']

                if _formatted_sentence[_jk] in i['NN'] and _formatted_sentence[_jk + 1] in i['NN'] and _formatted_sentence[_jk + 2] in i['IN'] and _formatted_sentence[_jk + 3] in i['NN']:
                    _ques = {}
                    _vb = copy.deepcopy(_formatted_sentence)

                    _vb[_jk] = "_____"
                    _vb[_jk + 1] = "_____"
                    _ques['question'] = " ".join(_vb)
                    _ques['option1'] = _formatted_sentence[_jk]
                    _ques['option2'] = _formatted_sentence[_jk + 1]

                    _ar = random.choice(i['NN'])
                    while _ar in [_formatted_sentence[_jk], _formatted_sentence[_jk + 1]]:
                        _ar = random.choice(i['NN'])

                    _ques['option3'] = _ar
                    _ques['option4'] = "None of the above"
                    _ques['answer1'] = _ques['option1']
                    _ques['answer2'] = _ques['option2']

                    _ques['score'] = 2
                    _ques['type'] = 1
                    self.question.append(_ques)


    def __generate_question(self):
        """
        Looks for the following patters:

        {<JJ><JJ><JJ>}
        {<JJ><CC><JJ>
        {<JJ><JJ>}
        {<NN><NN>}
        """


        for i in self.processed_transcript['processed-sentences']:
            _formatted_sentence = self.__format_string(i['sentence']).split()
            _temp_list = []

            if 'JJ' in i:
                for j in i['JJ']:
                    if j in _formatted_sentence:
                        _temp_list.append(_formatted_sentence.index(j))

            if 'CC' not in i:
                i['CC'] = []

            _temp_list.sort()

            _to_replace = []
            for k in range(1, len(_temp_list)):
                if _temp_list[k] == _temp_list[k-1] + 1:
                    if _temp_list[k-1] == _temp_list[k-2] + 1:
                        _to_replace.append([_temp_list[k-2], _temp_list[k-1], _temp_list[k]])
                    else:
                        _to_replace.append([_temp_list[k-1], _temp_list[k]])
                elif _temp_list[k] == _temp_list[k-1] + 2:
                    if _formatted_sentence[_temp_list[k-1]+1] in i['CC']:
                        _to_replace.append([_temp_list[k], _temp_list[k-1]])

            for kl in _to_replace:
                _ques = {}
                _vb = copy.deepcopy(_formatted_sentence)
                _rest_option = i['JJ']

                _igh = 1
                for _gh in kl:
                    _ques['option' + str(_igh)] = _formatted_sentence[_gh]
                    if _formatted_sentence[_gh] in _rest_option:
                        _rest_option.remove(_formatted_sentence[_gh])
                    _ques['answer' + str(_igh)] = _formatted_sentence[_gh]
                    _vb[_gh] = "_____"
                    _igh += 1

                _ques['question'] = " ".join(_vb)

                if _igh == 3 or _igh == 4:
                    for _jk in i['JJ']:
                        if _jk != _ques['option1'] and _jk != _ques['option2']:
                            _ques['option' + str(_igh)] = _jk
                            _igh += 1
                        if _igh == 5:
                            break

                if _igh == 4:
                    _ques['option4'] = "None of these"

                if _igh == 3:
                    _extra_list = []
                    if 'NN' in i:
                        _extra_list += i['NN']
                    if 'NNP' in i:
                        _extra_list += i['NNP']
                    if 'NNS' in i:
                        _extra_list += i['NNS']
                    _ques['option3'] = random.choice(_extra_list)
                    _ques['option4'] = "None of these"

                _ques['score'] = len(_ques) - 5
                _ques['type'] = 3
                self.question.append(_ques)



    def questions(self):
        """
        Returns question in the format
        {
        question: <>,
        option1: <>,
        option2: <>,
        option3: <>,
        option4: <>,
        answer1: <>,
        answer2: <>,
        answer3: <>,
        score: <len(answer)>
        type: 3
        }

        """
        self.__generate_question_noun()
        self.__generate_question()
        return self.question



import random


class FillInTheBlanks:
    """
    Fill in the blanks

    Class to generate fill in the blanks type questions based on processed
    transcript. Replacing the proper nouns.

    Arguments:
        processed_transcript: dict;
    """

    def __init__(self, processed_transcript: dict):
        """
        __init__

        Arguments:
            processed_transcript: dict
        """
        self.processed_transcript = processed_transcript

    def __generate_options(self):
        self.options = set()
        for i in self.processed_transcript["processed-sentences"]:
            if 'NNP' in i:
                self.options.update(set(i['NNP']))
            if 'NN' in i:
                self.options.update(set(i['NN']))

    def __generate_question(self):
        """
        generate question


        Option are the list of proper noun and nouns from the
        processed_transcript. Question is generate by replacing the proper noun
        with a blank, and putting other nouns as options.
        """

        self.question = []

        for i in self.processed_transcript["processed-sentences"]:
            if 'NNP' in i and len(
                    self.options) >= 3 and i['NNP'][0] in self.options:
                _qu = {}
                _qu["type"] = 1
                _qu["question"] = i['sentence'].replace(i['NNP'][0], "_____")
                _qu["option1"] = i["NNP"][0]

                self.options.remove(i["NNP"][0])
                _qu["option2"] = random.sample(self.options, 1)[0]

                self.options.remove(_qu["option2"])
                _qu["option3"] = random.sample(self.options, 1)[0]

                self.options.remove(_qu["option3"])
                _qu["option4"] = "None of the above"

                if len(self.options) >= 1:
                    _qu["option4"] = random.sample(self.options, 1)[0]

                _qu["answer"] = _qu["option1"]
                _qu["score"] = 1

                self.question.append(_qu)

    def questions(self):
        """
        Return List of FITB question:
            {
                "type": int
                "question": str,
                "option1": str,
                "option2": str,
                "option3": str,
                "option4": str
                "answer": str,
                "score": int
            }
        """
        self.__generate_options()
        self.__generate_question()

        return self.question



from setuptools import setup, find_packages

with open("README.md", "r") as readme_file:
    readme = readme_file.read()

requirements = ["urllib3==1.25.8", "beautifulsoup4==4.8.2",  "numpy==1.19.0", "cython==0.29.20", "spacy==2.3.2", "nltk==3.2.5", "language-tool-python==2.4.7", "language-tool==0.3", "textblob==0.15.3", "requests==2.22.0", "rake-nltk==1.0.4", "word2number==1.1"]

setup(
    name="paradigm-brain",
    version="0.0.33",
    author="Vedang Joshi",
    author_email="vedangj044@gmail.com",
    description="A package provides the core implementation for various use cases of paradigm.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/Paradigm-shift-AI/paradigm-brain/",
    packages=find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License"],
)



from textblob import TextBlob

class GivePOSTags:
    """
    GivePOSTags

    Class that adds the part of speech tags to the each sentence in the list
    of sentences.

    Arguments:
        list_of_sentences: List
    """

    def __init__(self, list_of_sentences: list):
        """
        __init__

        Arguments:
            list_of_sentences: List
        """

        self.list_of_sentences = list_of_sentences
        self.sentences_with_tags = []

    def __add_tags(self):
        for sentence in self.list_of_sentences:
            _temp = {}
            _temp["sentence"] = sentence
            _tags = TextBlob(sentence).tags

            for tg in _tags:
                if tg[1] not in _temp:
                    _temp[tg[1]] = []
                _temp[tg[1]].append(tg[0])

            self.sentences_with_tags.append(_temp)

    def get_sentences_with_tags(self):
        """
        get_sentences_with_tags

        Returns:
            Dictionary: The schema is ->
                        {"sentence": <sentence>, "NNP": <NNP>}
                        All tags of the sentence in Key-Value pair
        """

        self.__add_tags()

        return self.sentences_with_tags
